// -*-C++-*-
//
// File:      Task.cc
// Author:    jbakker
// Generated: Mon Mar 27 11:59:35 MET DST 2000

#include "Task.h"
#include "Register.h"

#include <SasException.h>

#include <STime.h>
#include <StopWatch.h>

#include <list>
#include <iostream>

#include <cstdio>
#include <strings.h>

using namespace Info;

Task* Task::_current(0);

//>>> invisible to the outside world!

/*
  envMetaName=name of environment
  envMetaLine="name=value"
 */
const char* envMetaName("__SAS_META_TASK__");
char        envMetaLine[2048];


/*
  We may change this to a map if we want to lookup faster.
  We won't gain much though, 'cause:
  -> how many tasks will a metatask query?
  -> how do we look it up, when using 'old' approach "task(mytask);"?
 */
class TaskList : public list<Task* >
{
public:
  static TaskList* instance();
private:
  TaskList(){};
  static TaskList* _instance;
};
TaskList* TaskList::_instance=0;
TaskList* TaskList::instance()
{
  if (_instance==0) {_instance=new TaskList;}
  return _instance;
}


/*
  The ServerRegistry keeps track of what servers need registration of this task
 */
namespace Info
{
class ServerRegistry : public list<Register* > {};
};

//<<< invisible to the outside world!

Task::Task(const char* name,const char* version, 
	   const char*  release, const char* package, Routine method)
  : _name(name),_version(version),_release(release),_package(package),
    _param(0),_global(0),_method(method), _registry(0)
{

  // add it to the list of Tasks
  TaskList::instance()->push_back(this);

}

Task::~Task()
{
  /*TODO: fix the bug in registry destruction!
  if ( _registry ) {
    for (ServerRegistry::iterator it=_registry->begin();
	 it != _registry->end();it++) {
      delete (*it);
      (*it)=0;
    }
    delete _registry;
  }
  for ( TaskList::iterator it= TaskList::instance()->begin();
      it != TaskList::instance()->end(); it++) {
    if ( (*it) == this ) (*it)=0;
  }
  */
}

void Task::exec(bool invoke)
{
  // keep:
  Task*          meta=_current;
  ParameterInfo* paramLast=gParameters; // last visited!

  // set the current task, so that it may query itself
  _current=this;

  // propagate parameters
  ::openParameters(_name);
  ParameterInfo* paramNow=gParameters;
  
  // don't do name mangling if called by main!
  bool byMain=(meta==0);
  if (byMain) invoke=false;

  errstr << "Executing " 
	 << (invoke?"(invoked): ":"(routine): ")
	 << paramNow->asCommandLine() << ' ' << errHandler.asCommandLine()
	 << Sas::message(Msg::UI,Msg::Silent);

  // name (package-version) [release]
  string taskStr(string(_name)    + " (" + string(_package) + "-" +
		 string(_version) + ") " + string(" [") + 
		 string(_release) + string("]"));

  // Clock the CPU load. We don't autostart, as we want to be as close
  // as we can to the real call/invocation
  StopWatch clock(StopWatch::CPU,false);

  errstr << taskStr << " started:  " << STime("now") 
	 << Sas::message(Msg::Meta, Msg::Verbose);

  // propagate parent's name, parameters etc...
  subscribe();
  if (invoke) {
    // set a hidden env variable, to pass on metatask name...
    sprintf(envMetaLine, "%s=%s",envMetaName,errHandler.name().c_str());
    putenv(envMetaLine);
    
    string exe(paramNow->asCommandLine() + " " + errHandler.asCommandLine());
    clock.start();
    if (system(exe.c_str())) { // FAILURE! cleanup and propagate error!

      // revert environment
      sprintf(envMetaLine, "%s=",envMetaName); putenv(envMetaLine);
      unsubscribe();
      _current=meta;
      gParameters=paramLast;

      errstr << "Execution failed of " << taskStr
	     << Sas::error("invalidInvocation");
    }
    clock.stop();
  }
  // propagate Task to error handler and execute
  else {
    string parentName=errHandler.name();
    if (! byMain) errHandler.name(errHandler.name() + "::" + _name);
    clock.start();

    // on FAILURE, cleanup and propagate!
    try {
      _method();
    }
    catch (SasException::Error e) {
      errHandler.name(parentName);
      unsubscribe();
      _current=meta;
      gParameters=paramLast;
      throw e;
    }
    catch (SasException::Fatal e) {
      errHandler.name(parentName);
      unsubscribe();
      _current=meta;
      gParameters=paramLast;
      throw e;
    }
    catch (SasException::Exception e) { // last SAS resort
      errHandler.name(parentName);
      unsubscribe();
      _current=meta;
      gParameters=paramLast;
      throw e;
    }
    catch (...) { // LAST RESORT!
      errHandler.name(parentName);
      unsubscribe();
      _current=meta;
      gParameters=paramLast;
      throw SasException::Fatal(meta->name(),"UnresolvedError",
				"Trapped unknown error");
    }

    clock.stop();
    errHandler.name(parentName);
  }
  unsubscribe();

  errstr << taskStr << " ended:    " << STime("now")
	 << Sas::message(Msg::Meta, Msg::Verbose);

  errstr << taskStr << " CPU time: " << clock.elapsed() << 's'
	 << Sas::message(Msg::Meta, Msg::Noisy);

  gParameters=paramLast;
  _current=meta;
}



Task* task(const char* name)
{
  Task* task=0;
  for ( TaskList::iterator it= TaskList::instance()->begin();
	it != TaskList::instance()->end(); it++) {
    if ( strcmp((*it)->name(),name) == 0 ) {
      task=(*it);
      break;
    }
  }
  return task;
}

void task(Routine routine)
{
  Task* task=0;
  for ( TaskList::iterator it= TaskList::instance()->begin();
	it != TaskList::instance()->end(); it++) {
    if ( (*it)->method() == routine ) {
      task=(*it);
      break;
    }
  }
  if (! task) errstr << "Task does not exist!" << Sas::error("invalidTask");

  task->exec();
}

//  WARNING: use only in 'setParameter()' and 'parameters()'
void Task::paramToGlobal()
{
  gParameters=_global;
}

//  WARNING: use only in 'setParameter()' and 'parameters()'
/*   
  Whenever "::openParameters()" is called, it sets the global ptr gParameters.
  In a meta task we don't want to screw with the wrong gParameters, so we
  kind of swap it temporarily...
 */
void Task::paramToTask()
{
  _global=gParameters; // current status-quo;

  // don't trust _param initially; make sure we have a valid pointer
  ::openParameters(_name);
  _param=gParameters;
}

/*
  make a Register available to this task, so we can subscribe to it, once 
  needed!
 */
void Task::append(Register* r)
{
  if (! _registry) _registry = new ServerRegistry;

   _registry->push_back(r);
}

/*
  subscribe to all servers that this task needs
 */
void Task::subscribe()
{
  if (! _registry) return;

    for (ServerRegistry::iterator it=_registry->begin();
	 it != _registry->end();it++) {
      (*it)->start();
    }
}

/*
   unsubscribe from all servers that this task needed
 */
void Task::unsubscribe()
{
  if (! _registry) return;

    for (ServerRegistry::iterator it=_registry->begin();
	 it != _registry->end();it++) {
      (*it)->stop();
    }
  
}
